"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import logging
import re
import socket
import time
import uuid
from collections import defaultdict
from collections.abc import Sequence
from datetime import timedelta
from typing import TYPE_CHECKING, Callable, Optional, Union

import importlib_metadata
from tornado import gen, queues, routing, web

import inmanta.protocol.endpoints
from inmanta.data.model import ExtensionStatus
from inmanta.protocol import Client, Result, TypedClient, common, endpoints, handle, methods
from inmanta.protocol.exceptions import ShutdownInProgress
from inmanta.protocol.rest import server
from inmanta.server import SLICE_SESSION_MANAGER, SLICE_TRANSPORT
from inmanta.server import config as opt
from inmanta.types import ArgumentTypes, JsonType
from inmanta.util import (
    CronSchedule,
    CycleException,
    IntervalSchedule,
    ScheduledTask,
    Scheduler,
    TaskHandler,
    TaskMethod,
    stable_depth_first,
)

if TYPE_CHECKING:
    from inmanta.server.extensions import Feature, FeatureManager

LOGGER = logging.getLogger(__name__)


class ServerStartFailure(Exception):
    pass


class SliceStartupException(ServerStartFailure):
    def __init__(self, slice_name: str, cause: Exception):
        super().__init__()
        self.__cause__ = cause
        self.in_slice = slice_name

    def __str__(self) -> str:
        return f"Slice {self.in_slice} failed to start because: {str(self.__cause__)}"


class ReturnClient(Client):
    """
    A client that uses a return channel to connect to its destination. This client is used by the server to communicate
    back to clients over the heartbeat channel.
    """

    def __init__(self, name: str, session: "Session") -> None:
        super().__init__(name, with_rest_client=False)
        self.session = session

    async def _call(
        self, method_properties: common.MethodProperties, args: list[object], kwargs: dict[str, object]
    ) -> common.Result:
        call_spec = method_properties.build_call(args, kwargs)
        expect_reply = method_properties.reply
        try:
            if method_properties.timeout:
                return_value = await self.session.put_call(
                    call_spec, timeout=method_properties.timeout, expect_reply=expect_reply
                )
            else:
                return_value = await self.session.put_call(call_spec, expect_reply=expect_reply)
        except asyncio.CancelledError:
            return common.Result(code=500, result={"message": "Call timed out"})

        return common.Result(code=return_value["code"], result=return_value["result"])


# Server Side
class Server(endpoints.Endpoint):
    def __init__(self, connection_timout: int = 120) -> None:
        super().__init__("server")
        self._slices: dict[str, ServerSlice] = {}
        self._slice_sequence: Optional[list[ServerSlice]] = None
        self._handlers: list[routing.Rule] = []
        self.connection_timout = connection_timout
        self.sessions_handler = SessionManager()
        self.add_slice(self.sessions_handler)

        self._transport = server.RESTServer(self.sessions_handler, self.id)
        self.add_slice(TransportSlice(self))
        self.running = False

    def add_slice(self, slice: "ServerSlice") -> None:
        """
        Add new endpoints to this rest transport
        """
        self._slices[slice.name] = slice
        self._slice_sequence = None

    def get_slices(self) -> dict[str, "ServerSlice"]:
        return self._slices

    def get_slice(self, name: str) -> "ServerSlice":
        return self._slices[name]

    def get_id(self) -> str:
        """
        Returns a unique id for a transport on an endpoint
        """
        return "server_rest_transport"

    id = property(get_id)

    def _order_slices(self) -> list["ServerSlice"]:
        edges: dict[str, set[str]] = defaultdict(set)

        for slice in self.get_slices().values():
            edges[slice.name].update(slice.get_dependencies())
            for depby in slice.get_depended_by():
                edges[depby].add(slice.name)

        names = list(edges.keys())
        try:
            order = stable_depth_first(names, {k: list(v) for k, v in edges.items()})
        except CycleException as e:
            raise ServerStartFailure("Dependency cycle between server slices " + ",".join(e.nodes)) from e

        def resolve(name: str) -> Optional["ServerSlice"]:
            if name in self._slices:
                return self._slices[name]
            LOGGER.debug("Slice %s is depended on but does not exist", name)
            return None

        return [s for s in (resolve(name) for name in order) if s is not None]

    def _get_slice_sequence(self) -> Sequence["ServerSlice"]:
        if self._slice_sequence is not None:
            return self._slice_sequence
        self._slice_sequence = self._order_slices()
        return self._slice_sequence

    async def start(self) -> None:
        """
        Start the transport.

        The order in which the different endpoints are prestarted/started, is determined by the
        order in which they are added to the RESTserver via the add_endpoint(endpoint) method.
        This order is hardcoded in the get_server_slices() method in server/bootloader.py
        """
        if self.running:
            return
        LOGGER.debug("Starting Server Rest Endpoint")
        self.running = True

        for my_slice in self._get_slice_sequence():
            try:
                LOGGER.debug("Pre Starting %s", my_slice.name)
                await my_slice.prestart(self)
            except Exception as e:
                raise SliceStartupException(my_slice.name, e)

        for my_slice in self._get_slice_sequence():
            try:
                LOGGER.debug("Starting %s", my_slice.name)
                await my_slice.start()
                self._handlers.extend(my_slice.get_handlers())
            except Exception as e:
                raise SliceStartupException(my_slice.name, e)

    async def stop(self) -> None:
        """
        Stop the transport.

        The order in which the endpoint are stopped, is reverse compared to the starting order.
        This prevents database connection from being closed too early. This order in which the endpoint
        are started, is hardcoded in the get_server_slices() method in server/bootloader.py
        """
        if not self.running:
            return
        self.running = False

        await super().stop()

        order = list(reversed(self._get_slice_sequence()))

        for endpoint in order:
            LOGGER.debug("Pre Stopping %s", endpoint.name)
            await endpoint.prestop()

        for endpoint in order:
            LOGGER.debug("Stopping %s", endpoint.name)
            await endpoint.stop()


class ServerSlice(inmanta.protocol.endpoints.CallTarget, TaskHandler[Result | None]):
    """
    Base class for server extensions offering zero or more api endpoints

    Extensions developers should override the lifecycle methods:

    * :func:`ServerSlice.prestart`
    * :func:`ServerSlice.start`
    * :func:`ServerSlice.prestop`
    * :func:`ServerSlice.stop`
    * :func:`ServerSlice.get_dependencies`

    To register endpoints that server static content, either use :func:'add_static_handler' or :func:'add_static_content'
    To create endpoints, use the annotation based mechanism

    To schedule recurring tasks, use :func:`schedule` or `self._sched`
    To schedule background tasks, use :func:`add_background_task`
    """

    feature_manager: "FeatureManager"

    def __init__(self, name: str) -> None:
        super().__init__()

        self._name: str = name
        self._handlers: list[routing.Rule] = []
        self._sched = Scheduler(f"server slice {name}")
        # is shutdown in progress?
        self._stopping: bool = False

    def is_stopping(self) -> bool:
        """True when prestop has been called."""
        return self._stopping

    async def prestart(self, server: Server) -> None:
        """
        Called by the RestServer host prior to start, can be used to collect references to other server slices
        Dependencies are not up yet.
        """

    async def start(self) -> None:
        """
        Start the server slice.

        This method `blocks` until the slice is ready to receive calls

        Dependencies are up (if present) prior to invocation of this call
        """

    async def prestop(self) -> None:
        """
        Always called before stop

        Stop producing new work:
        - stop timers
        - stop listeners
        - notify shutdown to systems depending on us (like agents)

        sets is_stopping to true

        But remain functional

        All dependencies are up (if present)
        """
        self._stopping = True
        await self._sched.stop()

    async def stop(self) -> None:
        """
        Go down

        All dependencies are up (if present)

        This method `blocks` until the slice is down
        """
        await super().stop()

    def get_dependencies(self) -> list[str]:
        """List of names of slices that must be started before this one."""
        return []

    def get_depended_by(self) -> list[str]:
        """List of names of slices that must be started after this one."""
        return []

    # internal API towards extension framework
    name = property(lambda self: self._name)

    def get_handlers(self) -> list[routing.Rule]:
        """Get the list of"""
        return self._handlers

    # utility methods for extensions developers
    def schedule(
        self,
        call: TaskMethod,
        interval: float = 60,
        initial_delay: Optional[float] = None,
        cancel_on_stop: bool = True,
        quiet_mode: bool = False,
    ) -> None:
        """
        Schedule a task repeatedly with a given interval. Tasks with the same call and the same schedule are considered the
        same. Clients that wish to be able to delete tasks should make sure to use a unique `call` function.

        :param interval: The interval between executions of the task.
        :param initial_delay: The delay to execute the task for the first time. If not set, interval is used.
        :quiet_mode: Set to true to disable logging the recurring notification that the action is being called. Use this to
        avoid polluting the server log for very frequent actions.
        """
        self._sched.add_action(call, IntervalSchedule(interval, initial_delay), cancel_on_stop, quiet_mode)

    def schedule_cron(self, call: TaskMethod, cron: str, cancel_on_stop: bool = True) -> None:
        """
        Schedule a task according to a cron specifier. Tasks with the same call and the same schedule are considered the same.
        Clients that wish to be able to delete tasks should make sure to use a unique `call` function.

        :param cron: The cron specifier to schedule the task by.
        """
        self._sched.add_action(call, CronSchedule(cron=cron), cancel_on_stop)

    def remove_cron(self, call: TaskMethod, cron: str) -> None:
        """
        Remove a cron-scheduled task.
        """
        self._sched.remove(ScheduledTask(action=call, schedule=CronSchedule(cron=cron)))

    def add_static_handler(self, location: str, path: str, default_filename: Optional[str] = None, start: bool = False) -> None:
        """
        Configure a static handler to serve data from the specified path.
        """
        if location[0] != "/":
            location = "/" + location

        if location[-1] != "/":
            location = location + "/"

        options = {"path": path}
        if default_filename is None:
            options["default_filename"] = "index.html"

        self._handlers.append(routing.Rule(routing.PathMatches(r"%s(.*)" % location), web.StaticFileHandler, options))
        self._handlers.append(
            routing.Rule(routing.PathMatches(r"%s" % location[:-1]), web.RedirectHandler, {"url": location[1:]})
        )
        if start:
            self._handlers.append((r"/", web.RedirectHandler, {"url": location[1:]}))

    def add_static_content(self, path: str, content: str, content_type: str = "application/javascript") -> None:
        self._handlers.append(
            routing.Rule(
                routing.PathMatches(r"%s(.*)" % path),
                server.StaticContentHandler,
                {"transport": self, "content": content, "content_type": content_type},
            )
        )

    def get_extension_status(self) -> Optional[ExtensionStatus]:
        ext_name = self.name.split(".")[0]
        source_package_name = self.__class__.__module__.split(".")[0]
        # workaround for #2586
        package_name = "inmanta-core" if source_package_name == "inmanta" else source_package_name
        try:
            distribution = importlib_metadata.distribution(package_name)
            return ExtensionStatus(name=ext_name, package=ext_name, version=distribution.version)
        except importlib_metadata.PackageNotFoundError:
            LOGGER.info(
                "Package %s of slice %s is not packaged in a distribution. Unable to determine its extension.",
                package_name,
                self.name,
            )
            return None

    @classmethod
    def get_extension_statuses(cls, slices: list["ServerSlice"]) -> list[ExtensionStatus]:
        result = {}
        for server_slice in slices:
            ext_status = server_slice.get_extension_status()
            if ext_status is not None:
                result[ext_status.name] = ext_status
        return list(result.values())

    async def get_status(self) -> dict[str, ArgumentTypes]:
        """
        Get the status of this slice.
        """
        return {}

    def define_features(self) -> list["Feature[object]"]:
        """Return a list of feature that this slice offers"""
        return []


class Session:
    """
    An environment that segments agents connected to the server. Should only be created in a context with a running event loop.
    """

    def __init__(
        self,
        sessionstore: "SessionManager",
        sid: uuid.UUID,
        hang_interval: int,
        timout: int,
        tid: uuid.UUID,
        endpoint_names: set[str],
        nodename: str,
        disable_expire_check: bool = False,
    ) -> None:
        self._sid = sid
        self._interval = hang_interval
        self._timeout = timout
        self._sessionstore: SessionManager = sessionstore
        self._seen: float = time.monotonic()
        self._callhandle: Optional[asyncio.TimerHandle] = None
        self.expired: bool = False

        self.last_dispatched_call: float = 0
        self.dispatch_delay = 0.01  # keep at least 10 ms between dispatches

        self.tid: uuid.UUID = tid
        self.endpoint_names: set[str] = endpoint_names
        self.nodename: str = nodename

        self._replies: dict[uuid.UUID, asyncio.Future] = {}

        # Disable expiry in certain tests
        if not disable_expire_check:
            self.check_expire()
        self._queue: queues.Queue[Optional[common.Request]] = queues.Queue()

        self.client = ReturnClient(str(sid), self)

    def check_expire(self) -> None:
        if self.expired:
            LOGGER.exception("Tried to expire session already expired")
        now = time.monotonic()
        ttw = self._timeout + self._seen - now
        if ttw < 0:
            expire_coroutine = self.expire(self._seen - now)
            self._sessionstore.add_background_task(expire_coroutine)
        else:
            self._callhandle = asyncio.get_running_loop().call_later(ttw, self.check_expire)

    def get_id(self) -> uuid.UUID:
        return self._sid

    id = property(get_id)

    async def expire(self, timeout: float) -> None:
        if self.expired:
            return
        self.expired = True
        if self._callhandle is not None:
            self._callhandle.cancel()
        await self._sessionstore.expire(self, timeout)

    def seen(self, endpoint_names: set[str]) -> None:
        self._seen = time.monotonic()
        self.endpoint_names = endpoint_names

    async def _handle_timeout(self, future: asyncio.Future, timeout: int, log_message: str) -> None:
        """A function that awaits a future until its value is ready or until timeout. When the call times out, a message is
        logged. The future itself will be cancelled.

        This method should be called as a background task. Any other exceptions (which should not occur) will be logged in
        the background task.
        """
        try:
            await asyncio.wait_for(future, timeout)
        except asyncio.TimeoutError:
            LOGGER.warning(log_message)

    def put_call(self, call_spec: common.Request, timeout: int = 10, expect_reply: bool = True) -> asyncio.Future:
        reply_id = uuid.uuid4()
        future = asyncio.Future()

        LOGGER.debug("Putting call %s: %s %s for agent %s in queue", reply_id, call_spec.method, call_spec.url, self._sid)

        if expect_reply:
            call_spec.reply_id = reply_id
            self._sessionstore.add_background_task(
                self._handle_timeout(
                    future,
                    timeout,
                    f"Call {reply_id}: {call_spec.method} {call_spec.url} for agent {self._sid} timed out.",
                )
            )
            self._replies[reply_id] = future
        else:
            future.set_result({"code": 200, "result": None})
        self._queue.put(call_spec)

        return future

    async def get_calls(self, no_hang: bool) -> Optional[list[common.Request]]:
        """
        Get all calls queued for a node. If no work is available, wait until timeout. This method returns none if a call
        fails.
        """
        try:
            call_list: list[common.Request] = []

            if no_hang:
                timeout = 0.1
            else:
                timeout = self._interval if self._interval > 0.1 else 0.1
                # We choose to have a minimum of 0.1 as timeout as this is also the value used for no_hang.
                # Furthermore, the timeout value cannot be zero as this causes an issue with Tornado:
                # https://github.com/tornadoweb/tornado/issues/3271
            call = await self._queue.get(timeout=timedelta(seconds=timeout))
            if call is None:
                # aborting session
                return None
            call_list.append(call)
            while self._queue.qsize() > 0:
                call = await self._queue.get()
                if call is None:
                    # aborting session
                    return None
                call_list.append(call)

            return call_list

        except gen.TimeoutError:
            return None

    def set_reply(self, reply_id: uuid.UUID, data: JsonType) -> None:
        LOGGER.log(3, "Received Reply: %s", reply_id)
        if reply_id in self._replies:
            future: asyncio.Future = self._replies[reply_id]
            del self._replies[reply_id]
            if not future.done():
                future.set_result(data)
        else:
            LOGGER.debug("Received Reply that is unknown: %s", reply_id)

    def get_client(self) -> ReturnClient:
        return self.client

    def abort(self) -> None:
        "Send poison pill to signal termination."
        self._queue.put(None)

    async def expire_and_abort(self, timeout: float) -> None:
        await self.expire(timeout)
        self.abort()


class SessionListener:
    async def new_session(self, session: Session, endpoint_names_snapshot: set[str]) -> None:
        """
        Notify that a new session was created.

        :param session: The session that was created
        :param endpoint_names_snapshot: The endpoint_names field of the session object may be updated after this
                                        method was called. This parameter provides a snapshot which will not change.
        """

    async def expire(self, session: Session, endpoint_names_snapshot: set[str]) -> None:
        """
        Notify that a session expired.

        :param session: The session that was created
        :param endpoint_names_snapshot: The endpoint_names field of the session object may be updated after this
                                        method was called. This parameter provides a snapshot which will not change.
        """

    async def seen(self, session: Session, endpoint_names_snapshot: set[str]) -> None:
        """
        Notify that a heartbeat was received for an existing session.

        :param session: The session that was created
        :param endpoint_names_snapshot: The endpoint_names field of the session object may be updated after this
                                        method was called. This parameter provides a snapshot which will not change.
        """


# Internals
class TransportSlice(ServerSlice):
    """Slice to manage the listening socket"""

    def __init__(self, server: Server) -> None:
        super().__init__(SLICE_TRANSPORT)
        self.server = server

    def get_dependencies(self) -> list[str]:
        """All Slices with an http endpoint should depend on this one using :func:`get_dependened_by`"""
        return []

    async def start(self) -> None:
        await super().start()
        await self.server._transport.start(self.server.get_slices().values(), self.server._handlers)

    async def prestop(self) -> None:
        await super().prestop()
        LOGGER.debug("Stopping Server Rest Endpoint")
        await self.server._transport.stop()

    async def stop(self) -> None:
        await super().stop()
        await self.server._transport.join()

    async def get_status(self) -> dict[str, ArgumentTypes]:
        def format_socket(sock: socket.socket) -> str:
            sname = sock.getsockname()
            return f"{sname[0]}:{sname[1]}"

        sockets = []
        if self.server._transport._http_server._sockets:
            sockets = [
                format_socket(s)
                for s in self.server._transport._http_server._sockets.values()
                if s.family in [socket.AF_INET, socket.AF_INET6]
            ]

        return {
            "inflight": self.server._transport.inflight_counter,
            "running": self.server._transport.running,
            "sockets": sockets,
        }


class SessionManager(ServerSlice):
    """
    A service that receives method calls over one or more transports
    """

    __methods__: dict[str, tuple[str, Callable]] = {}

    def __init__(self) -> None:
        super().__init__(SLICE_SESSION_MANAGER)

        # Config
        interval: int = opt.agent_timeout.get()
        hangtime: Optional[int] = opt.agent_hangtime.get()

        if hangtime is None:
            hangtime = int(interval * 3 / 4)

        self.hangtime: int = hangtime
        self.interval: int = interval

        # Session management
        self._sessions: dict[uuid.UUID, Session] = {}
        self._sessions_lock = asyncio.Lock()

        # Listeners
        self.listeners: list[SessionListener] = []

    async def get_status(self) -> dict[str, ArgumentTypes]:
        return {"hangtime": self.hangtime, "interval": self.interval, "sessions": len(self._sessions)}

    def add_listener(self, listener: SessionListener) -> None:
        self.listeners.append(listener)

    async def prestop(self) -> None:
        async with self._sessions_lock:
            # Keep the super call in the session_lock to make sure that no additional sessions are created
            # while the server is shutting down. This call sets the is_stopping() flag to true.
            await super().prestop()
        # terminate all sessions cleanly
        for session in self._sessions.copy().values():
            await session.expire(0)
            session.abort()

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    def validate_sid(self, sid: uuid.UUID) -> bool:
        if isinstance(sid, str):
            sid = uuid.UUID(sid)
        return sid in self._sessions

    async def get_or_create_session(self, sid: uuid.UUID, tid: uuid.UUID, endpoint_names: set[str], nodename: str) -> Session:
        if isinstance(sid, str):
            sid = uuid.UUID(sid)

        async with self._sessions_lock:
            if self.is_stopping():
                raise ShutdownInProgress()
            if sid not in self._sessions:
                session = self.new_session(sid, tid, endpoint_names, nodename)
                self._sessions[sid] = session
                endpoint_names_snapshot = set(session.endpoint_names)
                await asyncio.gather(*[listener.new_session(session, endpoint_names_snapshot) for listener in self.listeners])
            else:
                session = self._sessions[sid]
                self.seen(session, endpoint_names)
                endpoint_names_snapshot = set(session.endpoint_names)
                await asyncio.gather(*[listener.seen(session, endpoint_names_snapshot) for listener in self.listeners])

            return session

    def new_session(self, sid: uuid.UUID, tid: uuid.UUID, endpoint_names: set[str], nodename: str) -> Session:
        LOGGER.debug(f"New session with id {sid} on node {nodename} for env {tid} with endpoints {endpoint_names}")
        return Session(self, sid, self.hangtime, self.interval, tid, endpoint_names, nodename)

    async def expire(self, session: Session, timeout: float) -> None:
        async with self._sessions_lock:
            LOGGER.debug("Expired session with id %s, last seen %d seconds ago" % (session.get_id(), timeout))
            if session.id in self._sessions:
                del self._sessions[session.id]
            endpoint_names_snapshot = set(session.endpoint_names)
            await asyncio.gather(*[listener.expire(session, endpoint_names_snapshot) for listener in self.listeners])

    def seen(self, session: Session, endpoint_names: set[str]) -> None:
        LOGGER.debug("Seen session with id %s; endpoints: %s", session.get_id(), endpoint_names)
        session.seen(endpoint_names)

    @handle(methods.heartbeat, env="tid")
    async def heartbeat(
        self, sid: uuid.UUID, env: "inmanta.data.Environment", endpoint_names: list[str], nodename: str, no_hang: bool = False
    ) -> Union[int, tuple[int, dict[str, str]]]:
        LOGGER.debug("Received heartbeat from %s for agents %s in %s", nodename, ",".join(endpoint_names), env.id)

        session: Session = await self.get_or_create_session(sid, env.id, set(endpoint_names), nodename)

        LOGGER.debug("Let node %s wait for method calls to become available. (long poll)", nodename)

        # keep a minimal timeout between sending out calls to allow them to batch up
        now = time.monotonic()
        wait_time = session.dispatch_delay - (now - session.last_dispatched_call)
        if wait_time > 0:
            await asyncio.sleep(wait_time)

        call_list = await session.get_calls(no_hang=no_hang)

        if call_list is not None:
            LOGGER.debug("Pushing %d method calls to node %s", len(call_list), nodename)
            session.last_dispatched_call = time.monotonic()
            return 200, {"method_calls": call_list}
        else:
            LOGGER.debug("Heartbeat wait expired for %s, returning. (long poll)", nodename)

        return 200

    @handle(methods.heartbeat_reply)
    async def heartbeat_reply(
        self, sid: uuid.UUID, reply_id: uuid.UUID, data: JsonType
    ) -> Union[int, tuple[int, dict[str, str]]]:
        try:
            env = self._sessions[sid]
            env.set_reply(reply_id, data)
            return 200
        except Exception:
            LOGGER.warning(f"could not deliver agent reply with sid={sid} and reply_id={reply_id}", exc_info=True)
            return 500


class LocalClient(TypedClient):
    """A client that calls methods async on the server in the same process"""

    def __init__(self, name: str, server: Server) -> None:
        super().__init__(name, with_rest_client=False)
        self._server = server
        self._op_mapping: dict[str, dict[str, common.UrlMethod]] = {}
        for slice in server.get_slices().values():
            self._op_mapping.update(slice.get_op_mapping())

    def _get_op_mapping(self, url: str, method: str) -> common.UrlMethod:
        """Get the op mapping for the provided url and method"""
        methods = {}
        if url not in self._op_mapping:
            for key, mapping in self._op_mapping.items():
                if re.match(key, url):
                    methods = mapping
                    break
        else:
            methods = self._op_mapping[url]

        if method in methods:
            return methods[method]

        raise Exception(f"No handler defined for {method} {url}")

    async def _call(
        self, method_properties: common.MethodProperties, args: list[object], kwargs: dict[str, object]
    ) -> common.Result:
        spec = method_properties.build_call(args, kwargs)
        method_config = self._get_op_mapping(spec.url, spec.method)
        response = await self._server._transport._execute_call(method_config, spec.body, spec.headers)
        return self._process_response(method_properties, common.Result(code=response.status_code, result=response.body))



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from typing import Any, Optional

from tornado import web

from inmanta.types import JsonType


class BaseHttpException(web.HTTPError):
    """
    A base exception for errors in the server.

    Classes which extend from the BaseHttpException class cannot have mandatory arguments
    in their constructor. This is required to determine the status_code of the exception in
    :meth:`inmanta.protocol.common.MethodProperties._get_http_status_code_for_exception`
    """

    def __init__(self, status_code: int = 500, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        super().__init__(status_code, message)
        self.details = details

    def to_body(self) -> dict[str, Any]:
        """
        Return a response body
        """
        body: JsonType = {"message": self.log_message}
        if self.details is not None:
            body["error_details"] = self.details

        return body

    def to_status(self) -> int:
        """
        Return the status code
        """
        return self.status_code


class Forbidden(BaseHttpException):
    """
    An exception raised when access is denied (403)
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Access denied"
        if message is not None:
            msg += ": " + message

        super().__init__(403, msg, details)


class UnauthorizedException(BaseHttpException):
    """
    An exception raised when access to this resource is unauthorized
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Access to this resource is unauthorized"
        if message is not None:
            msg += ": " + message

        super().__init__(401, msg, details)


class BadRequest(BaseHttpException):
    """
    This exception is raised for a malformed request
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Invalid request"
        if message is not None:
            msg += ": " + message

        super().__init__(400, msg, details)


class NotFound(BaseHttpException):
    """
    This exception is used to indicate that a request or reference resource was not found.
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Request or referenced resource does not exist"
        if message is not None:
            msg += ": " + message

        super().__init__(404, msg, details)


class Conflict(BaseHttpException):
    """
    This exception is used to indicate that a request conflicts with the current state of the resource.
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Request conflicts with the current state of the resource"
        if message is not None:
            msg += ": " + message

        super().__init__(409, msg, details)


class ServerError(BaseHttpException):
    """
    An unexpected error occurred in the server
    """

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "An unexpected error occurred in the server while processing the request"
        if message is not None:
            msg += ": " + message

        super().__init__(500, msg, details)


class ShutdownInProgress(BaseHttpException):
    """This request can not be fulfilled because the server is going down"""

    def __init__(self, message: Optional[str] = None, details: Optional[JsonType] = None) -> None:
        msg = "Can not complete this request as a shutdown is on progress"
        if message is not None:
            msg += ": " + message

        super().__init__(503, msg, details)



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import datetime
import logging
import uuid
from collections.abc import Sequence
from typing import Optional, cast

from asyncpg import Connection

from inmanta import const, data
from inmanta.data import InvalidSort
from inmanta.data.dataview import NotificationsView
from inmanta.data.model import Notification
from inmanta.protocol import handle, methods_v2
from inmanta.protocol.common import ReturnValue
from inmanta.protocol.exceptions import BadRequest, NotFound
from inmanta.server import SLICE_COMPILER, SLICE_DATABASE, SLICE_NOTIFICATION, SLICE_TRANSPORT, protocol
from inmanta.server.services.compilerservice import CompilerService, CompileStateListener
from inmanta.server.validate_filter import InvalidFilter

LOGGER = logging.getLogger(__name__)


class NotificationService(protocol.ServerSlice, CompileStateListener):
    """Slice for notification management"""

    _compiler_service: CompilerService

    def __init__(self) -> None:
        super().__init__(SLICE_NOTIFICATION)

    def get_dependencies(self) -> list[str]:
        return [SLICE_DATABASE, SLICE_COMPILER]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        await super().prestart(server)
        self._compiler_service = cast(CompilerService, server.get_slice(SLICE_COMPILER))
        self._compiler_service.add_listener(self)

    async def start(self) -> None:
        await super().start()
        self.schedule(self._cleanup, 3600, initial_delay=0, cancel_on_stop=False)

    async def _cleanup(self) -> None:
        await data.Notification.clean_up_notifications()

    async def compile_done(self, compile: data.Compile) -> None:
        if not compile.success:
            compile_report = await data.Compile.get_report(compile_id=compile.id)
            reports = compile_report["reports"] if compile_report else []
            failed_pull_stage = next(
                (report for report in reports if report["name"] == "Pulling updates" and report["returncode"] != 0), None
            )
            if compile.notify_failed_compile is False:
                return
            elif compile.notify_failed_compile and compile.failed_compile_message:
                # Use specific message provided in request
                await self.notify(
                    compile.environment,
                    title="Compilation failed",
                    message=compile.failed_compile_message,
                    severity=const.NotificationSeverity.error,
                    uri=f"/api/v2/compilereport/{compile.id}",
                )
            elif failed_pull_stage:
                await self.notify(
                    compile.environment,
                    title="Pulling updates during compile failed",
                    message=failed_pull_stage["errstream"],
                    severity=const.NotificationSeverity.error,
                    uri=f"/api/v2/compilereport/{compile.id}",
                )
            elif compile.do_export:
                await self.notify(
                    compile.environment,
                    title="Compilation failed",
                    message="An exporting compile has failed",
                    severity=const.NotificationSeverity.error,
                    uri=f"/api/v2/compilereport/{compile.id}",
                )
            elif compile.notify_failed_compile:
                # Send notification with generic message as fallback
                await self.notify(
                    compile.environment,
                    title="Compilation failed",
                    message="A compile has failed",
                    severity=const.NotificationSeverity.error,
                    uri=f"/api/v2/compilereport/{compile.id}",
                )

    async def notify(
        self,
        environment: uuid.UUID,
        title: str,
        message: str,
        uri: Optional[str] = None,
        severity: const.NotificationSeverity = const.NotificationSeverity.message,
        connection: Optional[Connection] = None,
    ) -> None:
        """Internal API to create a new notification
        :param environment: The environment this notification belongs to
        :param title: The title of the notification
        :param message: The actual text of the notification
        :param severity: The severity of the notification
        :param uri: A link to an api endpoint of the server, that is relevant to the message,
                    and can be used to get further information about the problem.
                    For example a compile related problem should have the uri: `/api/v2/compilereport/<compile_id>`
        """
        await data.Notification(
            environment=environment,
            title=title,
            message=message,
            uri=uri,
            severity=severity,
            created=datetime.datetime.now().astimezone(),
        ).insert(connection)

    @handle(methods_v2.list_notifications, env="tid")
    async def list_notifications(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        first_id: Optional[uuid.UUID] = None,
        last_id: Optional[uuid.UUID] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "created.desc",
    ) -> ReturnValue[Sequence[Notification]]:
        try:
            handler = NotificationsView(
                environment=env,
                limit=limit,
                sort=sort,
                first_id=first_id,
                last_id=last_id,
                start=start,
                end=end,
                filter=filter,
            )
            out = await handler.execute()
            return out
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.get_notification, env="tid")
    async def get_notification(
        self,
        env: data.Environment,
        notification_id: uuid.UUID,
    ) -> Notification:
        notification = await data.Notification.get_one(environment=env.id, id=notification_id)
        if not notification:
            raise NotFound(f"Notification with id {notification_id} not found")
        return notification.to_dto()

    @handle(methods_v2.update_notification, env="tid")
    async def update_notification(
        self,
        env: data.Environment,
        notification_id: uuid.UUID,
        read: Optional[bool] = None,
        cleared: Optional[bool] = None,
    ) -> Notification:
        notification = await data.Notification.get_one(environment=env.id, id=notification_id)
        if not notification:
            raise NotFound(f"Notification with id {notification_id} not found")
        if read is not None and cleared is not None:
            await notification.update(read=read, cleared=cleared)
        elif read is not None:
            await notification.update(read=read)
        elif cleared is not None:
            await notification.update(cleared=cleared)
        else:
            raise BadRequest("At least one of {read, cleared} should be specified for a valid update")
        return notification.to_dto()



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os.path
import sys
from io import StringIO
from typing import Optional

import pytest
import yaml

import inmanta
from inmanta import config
from inmanta.logging import InmantaLoggerConfig, MultiLineFormatter, Options


@pytest.fixture(autouse=True)
def cleanup_logger():
    InmantaLoggerConfig.clean_instance()


def test_setup_instance():
    inmanta_logger = InmantaLoggerConfig.get_instance()
    handler = inmanta_logger.get_handler()
    assert handler.stream == sys.stdout
    assert isinstance(handler.formatter, MultiLineFormatter)
    assert handler.level == logging.INFO


def test_setup_instance_2_times():
    inmanta_logger = InmantaLoggerConfig.get_instance(sys.stderr)
    handler = inmanta_logger.get_handler()
    assert handler.stream == sys.stderr
    assert isinstance(handler.formatter, MultiLineFormatter)
    assert handler.level == logging.INFO

    inmanta_logger = InmantaLoggerConfig.get_instance(sys.stderr)
    handler = inmanta_logger.get_handler()
    assert handler.stream == sys.stderr
    assert isinstance(handler.formatter, MultiLineFormatter)
    assert handler.level == logging.INFO

    with pytest.raises(Exception) as e:
        InmantaLoggerConfig.get_instance(sys.stdout)
    message = "Instance already exists with a different stream"
    assert message in str(e.value)


def test_setup_instance_with_stream():
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    handler = inmanta_logger.get_handler()
    assert handler.stream == stream
    assert isinstance(handler.formatter, MultiLineFormatter)
    assert handler.level == logging.INFO

    # Log a message
    logger = logging.getLogger("test_logger")
    logger.info("This is a test message")
    log_output = stream.getvalue().strip()
    expected_output = "test_logger              INFO    This is a test message"
    assert log_output == expected_output


def test_set_log_level():
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    handler = inmanta_logger.get_handler()
    assert handler.level == logging.INFO

    logger = logging.getLogger("test_logger")
    expected_output = "test_logger              DEBUG   This is a test message"

    # Log a message and verify that it is not logged as the log level is too high
    logger.debug("This is a test message")
    log_output = stream.getvalue().strip()
    assert expected_output not in log_output

    # change the log_level and verify the log is visible this time.
    inmanta_logger.set_log_level("DEBUG")
    logger.debug("This is a test message")
    log_output = stream.getvalue().strip()
    assert expected_output in log_output


def test_set_log_formatter():
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)

    handler = inmanta_logger.get_handler()
    assert isinstance(handler.formatter, MultiLineFormatter)

    logger = logging.getLogger("test_logger")
    expected_output_format1 = "test_logger              INFO    This is a test message"
    expected_output_format2 = "test_logger - INFO - This is a test message"

    # Log a message with the default formatter
    logger.info("This is a test message")
    log_output = stream.getvalue().strip()
    assert expected_output_format1 in log_output

    # change the formatter and verify the output is different
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    inmanta_logger.set_log_formatter(formatter)
    assert inmanta_logger.get_handler().formatter == formatter

    logger.info("This is a test message")
    log_output = stream.getvalue().strip()
    assert expected_output_format2 in log_output


def test_set_logfile_location(
    tmpdir,
):
    log_file = tmpdir.join("test.log")
    inmanta_logger = InmantaLoggerConfig.get_instance()
    inmanta_logger.set_logfile_location(str(log_file))
    handler = inmanta_logger.get_handler()
    assert isinstance(handler, logging.handlers.WatchedFileHandler)
    assert handler.baseFilename == str(log_file)

    # Log a message
    logger = logging.getLogger("test_logger")
    logger.info("This is a test message")

    # Verify the message was written to the log file
    with open(str(log_file)) as f:
        contents = f.read()
        assert "This is a test message" in contents


@pytest.mark.parametrize_any(
    "log_file, log_file_level, verbose",
    [(None, "INFO", 1), (None, "ERROR", 4), ("test.log", "WARNING", 4), ("test.log", "DEBUG", 4)],
)
def test_apply_options(tmpdir, log_file, log_file_level, verbose):
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    logger = logging.getLogger("test_logger")

    if log_file:
        log_file = tmpdir.join("test.log")

    options1 = Options(log_file=log_file, log_file_level=log_file_level, verbose=verbose)
    inmanta_logger.apply_options(options1)
    logger.debug("debug: This is the first test")
    logger.info("info: This is the second test")
    logger.warning("warning: This is the third test")
    logger.error("error: This is the fourth test")
    if not log_file:
        log_output = stream.getvalue().strip()
        debug_in_output = "test_logger              DEBUG   debug: This is the first test" in log_output
        info_in_output = "test_logger              INFO    info: This is the second test" in log_output
        warning_in_output = "test_logger              WARNING warning: This is the third test" in log_output
        error_in_output = "test_logger              ERROR   error: This is the fourth test" in log_output
        assert debug_in_output if int(verbose) >= 3 else not debug_in_output
        assert info_in_output if int(verbose) >= 2 else not info_in_output
        assert warning_in_output if int(verbose) >= 1 else not warning_in_output
        assert error_in_output

    else:
        with open(str(log_file)) as f:
            log_output = f.read().strip()
            debug_in_output = "DEBUG    test_logger debug: This is the first test" in log_output
            info_in_output = "INFO     test_logger info: This is the second test" in log_output
            warning_in_output = "WARNING  test_logger warning: This is the third test" in log_output
            error_in_output = "ERROR    test_logger error: This is the fourth test" in log_output
            assert debug_in_output if log_file_level in ["DEBUG"] else not debug_in_output
            assert info_in_output if log_file_level in ["DEBUG", "INFO"] else not info_in_output
            assert warning_in_output if log_file_level in ["WARNING", "INFO", "DEBUG"] else not warning_in_output
            assert error_in_output


def test_logging_apply_options_2_times():
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    options1 = Options(log_file=None, log_file_level="INFO", verbose="1")
    inmanta_logger.apply_options(options1)
    with pytest.raises(Exception) as e:
        options2 = Options(log_file=None, log_file_level="INFO", verbose="2")
        inmanta_logger.apply_options(options2)
    message = "Options can only be applied once to a handler."
    assert message in str(e.value)


def test_logging_cleaned_after_apply_options(tmpdir):
    # verifies that when changing the stream with apply_option, the old stream is properly cleaned up
    # and not used anymore.
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    logger = logging.getLogger("test_logger")

    logger.info("This is a test message")
    log_output = stream.getvalue().strip()
    expected_output = "test_logger              INFO    This is a test message"
    assert log_output == expected_output

    log_file = tmpdir.join("test.log")

    options3 = Options(log_file=log_file, log_file_level="WARNING", verbose="4")
    inmanta_logger.apply_options(options3)
    logger.warning("warning: This is the second test")
    with open(str(log_file)) as f:
        contents = f.read()
        assert "WARNING  test_logger warning: This is the second test" in contents

    log_output = stream.getvalue().strip()
    expected_output = "test_logger              INFO    This is a test message"
    assert log_output == expected_output


def test_handling_logging_config_option(tmpdir, monkeypatch) -> None:
    """
    Verify the behavior of the logging_config option.
    """
    logger = logging.getLogger("TEST")

    stream = StringIO()
    # In order to reference an object in the logging_config file, it needs to be part of a module.
    # For this reason we add the 'pytest_stream' attribute to the inmanta module. It's referenced in the
    # logging_config file using ext://inmanta.pytest_stream
    monkeypatch.setattr(inmanta, "pytest_stream", stream, raising=False)

    def write_logging_config_file(path: str, formatter: str) -> None:
        config = {
            "version": 1,
            "formatters": {
                "console_formatter": {
                    "format": formatter,
                }
            },
            "handlers": {
                "console_handler": {
                    "class": "logging.StreamHandler",
                    "formatter": "console_formatter",
                    "level": "INFO",
                    "stream": "ext://inmanta.pytest_stream",
                },
            },
            "root": {
                "level": "INFO",
                "handlers": ["console_handler"],
            },
            "disable_existing_loggers": False,
        }
        with open(path, "w") as fh:
            yaml.dump(config, fh)

    def setup_logging_config(cli_options: Options, file_option_value: Optional[str] = None) -> None:
        # Set/Reset config options
        config.Config._reset()
        if file_option_value:
            config.logging_config.set(file_option_value)
        # Reset/Configure logging framework
        InmantaLoggerConfig.clean_instance()
        inmanta_logger_config = InmantaLoggerConfig.get_instance()
        inmanta_logger_config.apply_options(cli_options)
        # Reset stream buffer
        stream.truncate()

    path_logging_config_file1 = os.path.join(tmpdir, "logging_config1.yml")
    path_logging_config_file2 = os.path.join(tmpdir, "logging_config2.yml")
    path_logging_config_file3 = os.path.join(tmpdir, "logging_config3.yml")

    # Set --logging-config option on CLI only
    write_logging_config_file(path=path_logging_config_file1, formatter="AAA %(message)s")
    # Also assert that the other config options are ignored when logging_config is set.
    setup_logging_config(cli_options=Options(logging_config=path_logging_config_file1, verbose=1), file_option_value=None)
    logger.info("test")
    assert "AAA test" in stream.getvalue()

    # Set logging_config option in cfg file only
    write_logging_config_file(path=path_logging_config_file1, formatter="BBB %(message)s")
    setup_logging_config(cli_options=Options(), file_option_value=path_logging_config_file1)
    logger.info("test")
    assert "BBB test" in stream.getvalue()

    # Set the logging-config config option both on CLI and cfg file. CLI option takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="CCC %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="DDD %(message)s")
    setup_logging_config(
        cli_options=Options(logging_config=path_logging_config_file1),
        file_option_value=path_logging_config_file2,
    )
    logger.info("test")
    assert "CCC test" in stream.getvalue()

    # Set the logging-config config option in the cfg config file and using the environment variable.
    # The environment variable takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="EEE %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="FFF %(message)s")
    with monkeypatch.context() as m:
        m.setenv("INMANTA_CONFIG_LOGGING_CONFIG", path_logging_config_file1)
        setup_logging_config(
            cli_options=Options(),
            file_option_value=path_logging_config_file2,
        )
        logger.info("test")
        assert "EEE test" in stream.getvalue()

    # Set the logging-config config option on the CLI, in the cfg config file and using the environment variable.
    # The CLI option takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="GGG %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="HHH %(message)s")
    write_logging_config_file(path=path_logging_config_file3, formatter="III %(message)s")
    with monkeypatch.context() as m:
        m.setenv("INMANTA_CONFIG_LOGGING_CONFIG", path_logging_config_file1)
        setup_logging_config(
            cli_options=Options(logging_config=path_logging_config_file2),
            file_option_value=path_logging_config_file3,
        )
        logger.info("test")
        assert "HHH test" in stream.getvalue()



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import abc
import itertools
import logging
import re
from collections.abc import Sequence
from typing import Optional, TypeGuard, TypeVar, Union, overload

from inmanta.stable_api import stable_api

LOGGER = logging.getLogger(__name__)

TWDP = TypeVar("TWDP", bound="WildDictPath")
TWID = TypeVar("TWID", bound="WildInDict")
TWKL = TypeVar("TWKL", bound="WildKeyedList")
TWCP = TypeVar("TWCP", bound="WildComposedPath")


@stable_api
class InvalidPathException(Exception):
    """
    The path could not be parsed correctly.
    """


@stable_api
class ContainerStructureException(LookupError):
    """
    The requested item could not be found,
    because the container passed to this path is not of the expected type.
    """


@stable_api
class DictPathValue(abc.ABC):
    """
    Represents a data value part of a WildDictPath.
    """

    @abc.abstractmethod
    def escape(self) -> str:
        """
        Return this value with all special characters escaped.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def matches(self, value: Optional[object]) -> bool:
        """
        Return true iff the given value matches this value.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def value(self) -> Optional[str]:
        """
        Return the unescaped data value.
        """
        raise NotImplementedError()

    @classmethod
    def create(cls, value: str) -> "DictPathValue":
        """
        Create a DictPathValue from the given value. Argument `value` must have all
        special characters escaped when their special meaning is not desired.
        """
        if value == WildCardValue.WILDCARD_CHARACTER:
            return WildCardValue()
        elif value == NullValue.NULL_VALUE_CHARACTER:
            return NullValue()
        else:
            unescaped_value: str = WildDictPath.PATTERN_ESCAPED_SPECIAL_CHARACTER.sub(r"\1", value)
            return NormalValue(unescaped_value)

    @classmethod
    def from_object(cls, value: object) -> "DictPathValue":
        """
        Create a DictPathValue from the given object.
        :param value: The object to construct a DictPathValue for. It is interpreted as a literal value: if it is a string,
            special characters will not be interpreted and must not be escaped. If it is not `None`, it must implement `str()`
            to be an unambiguous representation of the object.
        """
        if value is None:
            return NullValue()
        return NormalValue(str(value))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)


@stable_api
class NormalValue(DictPathValue):
    """
    A normal dict path value. This is a data value that is matched literally
    against another value.
    """

    def __init__(self, value: str) -> None:
        super(DictPathValue, self).__init__()
        self._value: str = value
        self._numeric_value: Optional[float] = self._try_parse_numeric(value)

    def escape(self) -> str:
        return WildDictPath.PATTERN_SPECIAL_CHARACTER.sub(r"\\\1", self._value)

    def matches(self, value: Optional[object]) -> bool:
        if value is None:
            return False

        # Perform a numeric comparison only if the value is an int/float and
        # The key in the dictpath can be interpreted as an int/float
        if self._numeric_value is not None and isinstance(value, (int, float)):
            return self._numeric_value == value

        # Fallback to string comparison for other types
        return self._value == str(value)

    @staticmethod
    def _try_parse_numeric(value: str) -> Optional[float]:
        try:
            return float(value)
        except (ValueError, TypeError):
            return None

    @property
    def value(self) -> str:
        return self._value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NormalValue):
            return False
        return self.value == other.value


@stable_api
class WildCardValue(DictPathValue):
    """
    Represents a wildcard value. This is a data value that matches any other value.
    """

    WILDCARD_CHARACTER: str = "*"

    def escape(self) -> str:
        return self.WILDCARD_CHARACTER

    def matches(self, value: Optional[object]) -> bool:
        return True

    @property
    def value(self) -> str:
        raise Exception("A WildCardValue doesn't have an actual data value.")


@stable_api
class NullValue(DictPathValue):
    """
    Represents the data value None. Matches against any other None value.
    """

    NULL_VALUE_CHARACTER: str = r"\0"

    def escape(self) -> str:
        return self.NULL_VALUE_CHARACTER

    def matches(self, value: Optional[object]) -> bool:
        return value is None

    @property
    def value(self) -> None:
        return None


@stable_api
class WildDictPath(abc.ABC):
    """
    A base class for all dict paths segments.  It supports the usage of wildcards, allowing to reach
    multiple elements in the same Dict.

    The wildcard is the character "*", its usage for the different segment type is documented later.
    In the internal representation of this object, wildcards are represented using the value None.

    The special characters mentioned in `WildDictPath.SPECIAL_CHARACTERS` should be escaped using a
    backslash when its used as a data character (e.g. the value of a key in a dictionary) instead
    of a control character (a character used to guide the dictpath library). Other characters must
    never be escaped.
    """

    # Special characters should be escaped in data elements of the dict path
    # to prevent incorrect interpretation of the dict path.
    SPECIAL_CHARACTERS: list[str] = ["\\", "[", "]", ".", "*", "="]
    REGEX_SPECIAL_CHARACTER = rf"([{re.escape(''.join(SPECIAL_CHARACTERS))}])"
    PATTERN_SPECIAL_CHARACTER = re.compile(REGEX_SPECIAL_CHARACTER)
    REGEX_ESCAPED_SPECIAL_CHARACTER = rf"\\([{re.escape(''.join(SPECIAL_CHARACTERS))}])"
    PATTERN_ESCAPED_SPECIAL_CHARACTER = re.compile(REGEX_ESCAPED_SPECIAL_CHARACTER)
    REGEX_NORMAL_CHARACTER = rf"[^{re.escape(''.join(SPECIAL_CHARACTERS))}]"

    # Add the WILDCARD variable here for backwards compatibility
    WILDCARD = WildCardValue.WILDCARD_CHARACTER

    @abc.abstractmethod
    def get_elements(self, container: object) -> list[object]:
        """
        Get the elements identified by this Path from the given collection.
        If no element is matched, an empty list is returned.

        :param container: the container to search in
        """

    def resolve_wild_cards(self: TWDP, container: object) -> Sequence[TWDP]:
        """
        Resolve all the wild cards contained in this wild dict path, based on the
        given container.  For each possible set of values that can replace the wild
        cards and still resolve a subset of the elements matched by this expression,
        return a wild dict path expression without the use of any wild card.

        The set of all elements that can be resolved by all returned wild dict paths
        for the given container, should be equal to the set of all elements matched by
        this wild dict path expression for the given container.

        This method may return a NotImplementedError if the path upon which it is called
        isn't supported by the method.

        :param container: the container to search in
        """
        raise NotImplementedError(
            "There is no default behavior for resolve_wild_cards on the base class WildDictCard, "
            "it should be implemented in the sub classes."
        )

    @abc.abstractmethod
    def to_str(self) -> str:
        """
        Returns the dict path expression represented by this instance.
        """

    def __str__(self) -> str:
        return self.to_str()

    def __add__(self, other: object) -> "WildDictPath":
        if not isinstance(other, WildDictPath):
            return NotImplemented
        return WildComposedPath(path=list(self.get_path_sections()) + list(other.get_path_sections()))

    @abc.abstractmethod
    def get_path_sections(self) -> Sequence["WildDictPath"]:
        """
        A DictPath can be a combination of multiple DictPaths, this returns all the DictPaths
        that compose this one, or itself if it is not a composition of multiple DictPaths.
        """

    @classmethod
    @abc.abstractmethod
    def parse(cls: type[TWDP], inp: str) -> Optional[TWDP]:
        pass

    def _validate_container(self, container: object) -> TypeGuard[dict[object, object]]:
        return isinstance(container, dict)


@stable_api
class WildInDict(WildDictPath):
    """
    This is the path that, if you call get_element on a dict, it returns the value stored in
    that key in that dict. This class accepts only top level keys as its expression.

    The string representation of the following path element is `a`

    .. code_block:: python

        assert WildInDict("a").get_elements(
        {
            "a":"b",
            "c":"d",
        }) == ["b"]

    A wild card can be used to get all values from the dict.  The wildcard only works as a single character.

    .. code_block:: python

        assert WildInDict("*").get_elements(
            {
                "a":"b",
                "c":"d",
            }
        ) == ["b", "d"]

    The following code raises a KeyError.

    .. code_block:: python

        WildInDict("a*").get_elements(
            {
                "a":"b",
                "c":"d",
            }
        )

    """

    IN_DICT_PATTERN = re.compile(
        rf"(?P<key>({WildDictPath.REGEX_NORMAL_CHARACTER}|{WildDictPath.REGEX_ESCAPED_SPECIAL_CHARACTER})+|\*)"
    )

    def __init__(self, key: str) -> None:
        key_value = DictPathValue.create(key)
        if not isinstance(key_value, NormalValue) and not isinstance(key_value, WildCardValue):
            raise InvalidPathException(f"Invalid dictionary key {key}")
        self.key: Union[NormalValue, WildCardValue] = key_value

    def get_elements(self, container: object) -> list[object]:
        if self._validate_container(container):
            try:
                return [value for key, value in container.items() if self.key.matches(key)]
            except KeyError:
                return []
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def resolve_wild_cards(self, container: object) -> Sequence["WildInDict"]:
        """
        Get one WildInDict by key matching the wild key of this object.

        ..code-block:: python

            assert WildInDict("a").resolve_wild_cards(
                {
                    "a": "b",
                    "aa": "bb",
                    "c": "d",
                }
            ) == [
                WildInDict("a"),
            ]

            assert WildInDict("*").resolve_wild_cards(
                {
                    "a": "b",
                    "aa": "bb",
                    "c": "d",
                }
            ) == [
                WildInDict("a"),
                WildInDict("aa"),
                WildInDict("c"),
            ]

        """
        if self._validate_container(container):
            try:
                return [WildInDict(str(key)) for key in container if self.key.matches(key)]
            except KeyError:
                return []
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def to_str(self) -> str:
        return self.key.escape()

    def get_path_sections(self) -> Sequence[WildDictPath]:
        return [self]

    @classmethod
    def parse(cls: type[TWID], inp: str) -> Optional[TWID]:
        match = cls.IN_DICT_PATTERN.fullmatch(inp)
        if match:
            return cls(inp)
        return None


@stable_api
class WildKeyedList(WildDictPath):
    """
    Find a specific item in a list, based on a key-value pair.
    The list is in a dictionary itself.

    The string representation of the following path element is `relation[key_attribute=key_value]`
    A wild card can be used to get all values from the list having the key_attribute.

    e.g.::

        WildKeyedList("relation","key_attribute","key_value").get_elements(
        {
            "relation":[
                {
                    "key_attribute":"key_value",
                    "other_attribute":"other_value"
                },
                {
                    "key_attribute":"other_value"
                }
            [
        })

    will return::

        [
            {
                "key_attribute":"key_value",
                "other_attribute":"other_value"
            }
        ]

    e.g.::

        WildKeyedList("relation","key_attribute","*").get_elements(
        {
            "relation":[
                {
                    "key_attribute":"key_value",
                    "other_attribute":"other_value",
                },
                {
                    "key_attribute":"other_value",
                },
                {
                    "other_key_attribute":"other_value",
                },
            [
        })

    will return::

        [
            {
                "key_attribute":"key_value",
                "other_attribute":"other_value",
            },
            {
                "key_attribute":"other_value",
            }
        ]

    """

    REGEX_RELATION = rf"(?P<relation>({WildDictPath.REGEX_NORMAL_CHARACTER}|{WildDictPath.REGEX_ESCAPED_SPECIAL_CHARACTER})+)"
    REGEX_KEY_ATTRIBUTE = (
        rf"(?P<key_attribute>({WildDictPath.REGEX_NORMAL_CHARACTER}|{WildDictPath.REGEX_ESCAPED_SPECIAL_CHARACTER})+|\*)"
    )
    REGEX_KEY_VALUE = (
        rf"(?P<key_value>({WildDictPath.REGEX_NORMAL_CHARACTER}|{WildDictPath.REGEX_ESCAPED_SPECIAL_CHARACTER})*|\*|\\0)"
    )
    KEY_VALUE_PAIR = rf"\[{REGEX_KEY_ATTRIBUTE}={REGEX_KEY_VALUE}]"

    KEYED_LIST_PATTERN = re.compile(rf"^{REGEX_RELATION}(?P<selectors>({KEY_VALUE_PAIR})+)$")
    KEY_VALUE_PAIRS_PATTERN = re.compile(KEY_VALUE_PAIR)

    @overload
    def __init__(self, relation: str, key_value_pairs: Sequence[tuple[str, str]]) -> None:
        """
        :param relation: The relation on the object that is the keyed list
        :param key_value_pairs: The key-value pairs to look for in each item of the keyed list.
            The key is compared using string comparison. As such 5=="5" and False=="False"
        """
        ...

    @overload
    def __init__(self, relation: str, key_attribute: str, key_value: str, /) -> None:
        """
        Deprecated constructor, kept for backwards compatibility reasons.

        :param relation: The relation on the object that is the keyed list
        :param key_attribute: The attribute to look for in each item of the keyed list
        :param key_value: The attribute value to look for in each item of the keyed list.
        """

    def __init__(
        self, relation: str, key_value_pairs: Union[str, Sequence[tuple[str, str]]], key_value: Optional[str] = None
    ) -> None:
        if isinstance(key_value_pairs, str):
            LOGGER.warning(
                "The %s(relation: str, key_attribute: str, key_value: str, /) constructor is deprecated and will be removed"
                " in a future version. Please use %s(relation: str, key_value_pairs: Sequence[Tuple[str, str]]) instead",
                self.__class__.__name__,
                self.__class__.__name__,
            )
            assert key_value is not None
            key_value_pairs = [(key_value_pairs, key_value)]
        relation_value = DictPathValue.create(relation)
        if not isinstance(relation_value, NormalValue):
            raise InvalidPathException(f"Invalid relation name: {relation}")
        self.relation: NormalValue = relation_value

        if not key_value_pairs:
            raise ValueError("A keyed list path requires at least one key-value pair.")
        if len({pair[0] for pair in key_value_pairs}) != len(key_value_pairs):
            raise ValueError("No duplicate keys allowed in keyed list path")

        self.key_value_pairs: Sequence[tuple[Union[NormalValue, WildCardValue], DictPathValue]] = [
            (self._parse_key(pair[0]), self._parse_value(pair[1])) for pair in key_value_pairs
        ]

    @classmethod
    def _parse_key(cls, key: str) -> Union[NormalValue, WildCardValue]:
        """
        Parse a key string into the corresponding dict path object.
        """
        result: DictPathValue = DictPathValue.create(key)
        if not isinstance(result, NormalValue) and not isinstance(result, WildCardValue):
            raise InvalidPathException(f"Invalid dictionary key name: {key}")
        return result

    @classmethod
    def _parse_value(cls, value: str) -> DictPathValue:
        """
        Parse a value string into the corresponding dict path object.
        """
        return DictPathValue.create(value)

    def _validate_outer_container(self, container: object) -> dict[object, object]:
        if not isinstance(container, dict):
            raise ContainerStructureException(f"{container} is not a Dict")
        return container

    def _validate_inner_container(self, container: object) -> list[object]:
        if not isinstance(container, list):
            raise ContainerStructureException(f"{container} is not a List or Set")
        return container

    def get_elements(self, container: object) -> list[object]:
        outer = self._validate_outer_container(container)
        try:
            inner = outer[self.relation.value]
        except KeyError:
            return []
        the_list = self._validate_inner_container(inner)
        return [
            dct
            for dct in the_list
            if isinstance(dct, dict)
            and all(any(key.matches(k) and value.matches(v) for k, v in dct.items()) for key, value in self.key_value_pairs)
        ]

    def resolve_wild_cards(self, container: object) -> Sequence["WildKeyedList"]:
        """
        Get one WildKeyedList by set of matching key value pairs that can be found
        in the given container at the relation specified in this object.

        As of now, this method can not be called on dict path expressions using a wild card
        in the key of any key-value pair.  This is because the behavior for such type of path
        is undefined.  It might get defined later on, once we get a use case for them.
        If the method is called on such a path, a NotImplementedError is raised.

        ..code-block:: python

            assert WildKeyedList("relation","key_attribute","*").resolve_wild_cards(
                {
                    "relation":[
                        {
                            "key_attribute":"key_value",
                            "other_attribute":"other_value",
                        },
                        {
                            "key_attribute":"other_value",
                        },
                        {
                            "other_key_attribute":"other_value",
                        },
                    [
                }
            ) == [
                WildKeyedList("relation", "key_attribute", "key_value"),
                WildKeyedList("relation", "key_attribute", "other_value"),
            ]

            assert WildKeyedList("relation",[("key_attribute","*"), ("other_attribute", "*")]).resolve_wild_cards(
                {
                    "relation":[
                        {
                            "key_attribute":"key_value",
                            "other_attribute":"other_value",
                        },
                        {
                            "key_attribute":"other_value",
                        },
                        {
                            "other_key_attribute":"other_value",
                        },
                    [
                }
            ) == [
                WildKeyedList("relation", [("key_attribute","key_value"), ("other_attribute", "other_value")]),
            ]

        """
        # Check if we have a wild card for any of the keys, if we do, raise a NotImplementedError
        # as we didn't define an expected behavior for it yet
        for key, _ in self.key_value_pairs:
            if key == WildCardValue():
                raise NotImplementedError(
                    "Can not call resolve_wild_cards on this path because it uses "
                    f"wild cards (`*`) for some of its keys: `{self}`.  The desired "
                    "behavior of resolve_wild_cards for this type of path is currently undefined."
                )

        outer = self._validate_outer_container(container)
        try:
            inner = outer[self.relation.value]
        except KeyError:
            return []

        # Save here all the KeyedList that can be constructed
        matches: list[WildKeyedList] = []

        # Check each of the inner values, and emit as many paths for it
        # as can be created to match it (within the constraints of this path
        # filters)
        for dct in self._validate_inner_container(inner):
            if not isinstance(dct, dict):
                # Can't be a match, it is not even a dict
                continue

            # Save here all the key-value pairs that will compose our
            # KeyedList dict path
            pairs = [
                [(str(k), str(v)) for k, v in dct.items() if key.matches(k) and value.matches(v)]
                for key, value in self.key_value_pairs
            ]

            # Emit one KeyedList for each combination of matched key-value pairs
            matches.extend(
                WildKeyedList(
                    self.relation.value,
                    path,
                )
                for path in itertools.product(*pairs)
            )

        return matches

    def to_str(self) -> str:
        escaped_relation: str = self.relation.escape()
        escaped_key_value_pairs: str = "][".join(key.escape() + "=" + value.escape() for key, value in self.key_value_pairs)
        return f"{escaped_relation}[{escaped_key_value_pairs}]"

    def get_path_sections(self) -> Sequence[WildDictPath]:
        return [self]

    @classmethod
    def parse(cls: type[TWKL], inp: str) -> Optional[TWKL]:
        match = cls.KEYED_LIST_PATTERN.fullmatch(inp)
        if match:
            group_dct = match.groupdict()
            pairs: list[tuple[str, str]] = [
                (pair.group("key_attribute"), pair.group("key_value"))
                for pair in cls.KEY_VALUE_PAIRS_PATTERN.finditer(group_dct["selectors"])
            ]
            return cls(group_dct["relation"], pairs)
        return None

    def get_key_value_pairs(self) -> Sequence[tuple[str, Optional[str]]]:
        """
        Return a list of tuples, where each element in the list is a literal (unescaped) key-value pair for this WildKeyedList.
        """
        return [(key.value, value.value) for key, value in self.key_value_pairs]

    def __eq__(self, other: object) -> bool:
        if other.__class__ != self.__class__:
            return False
        assert isinstance(other, WildKeyedList)  # Make mypy happy
        return self.relation == other.relation and self.key_value_pairs == other.key_value_pairs


@stable_api
class WildComposedPath(WildDictPath):
    """
    A path composed of multiple elements, separated by "."
    """

    element_types: Sequence[type[WildDictPath]] = [WildInDict, WildKeyedList]
    COMPOSED_DICT_PATH_PATTERN = re.compile(r"(?:[^.\\]|\\.)+")

    def __init__(self, path_str: Optional[str] = None, path: Optional[Sequence[WildDictPath]] = None) -> None:
        if (path_str is None) == (path is None):
            raise ValueError("Either path or path_str should be set")

        self.path: str
        self.expanded_path: Sequence[WildDictPath]
        if path_str is not None:
            self.path = path_str
            self.expanded_path = self.do_parse(path_str)
        else:
            assert path is not None
            self.expanded_path = path
            self.path = self.un_parse()

    def un_parse(self) -> str:
        return ".".join(element.to_str() for element in self.expanded_path)

    @classmethod
    def split_on_dots(cls, path_str: str) -> list[str]:
        """
        Split the given `path_str` on dot characters if they are not escaped with a backslash.
        """
        match = cls.COMPOSED_DICT_PATH_PATTERN.findall(path_str)
        if not match:
            raise InvalidPathException(f"Could not parse path {path_str}")
        return match

    @classmethod
    def do_parse(cls, path_str: str) -> Sequence[WildDictPath]:
        splitted_path_str: list[str] = cls.split_on_dots(path_str)

        def parse_element(inp: str) -> WildDictPath:
            for subtype in cls.element_types:
                parsed = subtype.parse(inp)
                if parsed is not None:
                    return parsed
            raise InvalidPathException(f"Could not parse path segment {inp}")

        return [parse_element(e) for e in splitted_path_str]

    @classmethod
    def parse(cls: type[TWCP], inp: str) -> Optional[TWCP]:
        try:
            path = cls.do_parse(path_str=inp)
            return cls(path=path)
        except InvalidPathException:
            return None

    def get_elements(self, container: object) -> list[object]:
        if container is None:
            raise IndexError("Can not get anything from None")

        containers = [container]
        for item in self.expanded_path:
            next_containers = []
            for container in containers:
                next_containers.extend(item.get_elements(container))

            containers = next_containers

        return containers

    def resolve_wild_cards(self, container: object) -> Sequence["WildComposedPath"]:
        if container is None:
            raise IndexError("Can not get anything from None")

        if len(self.expanded_path) == 0:
            # No need to dig further
            return []

        if len(self.expanded_path) == 1:
            # The path is equivalent to its unique element, but we have to wrap
            # each of the possible path into a ComposedPath object to be consistent
            return [WildComposedPath(path=[path]) for path in self.expanded_path[0].resolve_wild_cards(container)]

        return [
            WildComposedPath(path=[path, *next_part.get_path_sections()])
            for path in self.expanded_path[0].resolve_wild_cards(container)
            for elem in path.get_elements(container)
            for next_part in WildComposedPath(path=self.expanded_path[1:]).resolve_wild_cards(elem)
        ]

    def to_str(self) -> str:
        return self.path

    def get_path_sections(self) -> Sequence[WildDictPath]:
        return self.expanded_path

    def __eq__(self, other: object) -> bool:
        if other.__class__ != self.__class__:
            return False
        assert isinstance(other, type(self)), f"{type(other)} != {type(self)}"  # Make mypy happy
        return self.expanded_path == other.expanded_path


@stable_api
class WildNullPath(WildDictPath):
    """
    A DictPath with no length

    (i.e. return the container itself, wrapped in a list)
    """

    def get_elements(self, container: object) -> list[object]:
        if self._validate_container(container):
            return [container]
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def resolve_wild_cards(self, container: object) -> Sequence["WildNullPath"]:
        if self._validate_container(container):
            return [WildNullPath()]
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def get_path_sections(self) -> Sequence["DictPath"]:
        return []

    def to_str(self) -> str:
        return "."

    @classmethod
    def parse(cls, inp: str) -> None:
        raise NotImplementedError("NullPath is not intended to be parseable, it should only be used programmatically.")


@stable_api
class DictPath(WildDictPath):
    """
    A base class for all non-wild dict paths segments.  The key difference between WildDictPath and DictPath subclasses are:
     1. WildDictPath can only get a list of elements, with get_elements.  If no element is found, an empty list is returned,
        no error is raised.
     2. DictPath can not use get_elements as it is always expected to have exactly one match.
     3. DictPath can use get_element, which will return the matching element, or raise an exception if more or less than one
        is found.
     4. DictPath can set values, using set_element, and can build the dict structure expected by the path by using the
        construct flag in the get_element method.
    """

    @abc.abstractmethod
    def get_element(self, container: object, construct: bool = False) -> object:
        """
        Get the element identified by this Path from the given collection

        :param container: the container to search in
        :param construct: construct a dict on the location identified by this path in the container
                          if the element doesn't exist. Return this new dict.

        :raises KeyError: if the element is not found or if more than one occurrence was found.
        """

    def get_elements(self, container: object) -> list[object]:
        try:
            return [self.get_element(container, False)]
        except LookupError:
            return []

    @abc.abstractmethod
    def set_element(self, container: object, value: object, construct: bool = True) -> None:
        """
        Set the element identified by this Path from the given collection.

        If construct is True, all containers on the path towards the value are constructed if absent.

        :raises LookupError: if the path leading to the element is not found or if more than one occurrence was found.
        """

    def __add__(self, other: object) -> "DictPath":
        if not isinstance(other, DictPath):
            return NotImplemented
        return ComposedPath(path=list(self.get_path_sections()) + list(other.get_path_sections()))

    def get_path_sections(self) -> Sequence["DictPath"]:
        """Get the individual parts of this path"""
        return []

    @abc.abstractmethod
    def get_key(self) -> str:
        """
        Return the dictionary key referenced by this element in the dict path.
        """

    @abc.abstractmethod
    def remove(self, container: object) -> None:
        """
        Remove an element if it exists:
            * On an InDict or a WildInDict: Remove the referenced key from the dictionary.
            * On a KeyedList or a WildKeyedList: Remove the referenced element from the list.
            * On a NullPath: This operation is not supported on a NullPath.
        """
        raise NotImplementedError()


@stable_api
class InDict(DictPath, WildInDict):
    """
    This is the path that, if you call get_element on a dict, it returns the value stored in that key in that dict.

    The string representation of the following path element is: "a"

    .. code_block:: python

        assert InDict("a").get_element(
        {
            "a":"b",
            "c":"d",
        }) == "b"

    """

    def __init__(self, key: str) -> None:
        WildInDict.__init__(self, key)
        if isinstance(self.key, WildCardValue):
            raise ValueError(f"The Wildcard ('{WildCardValue.WILDCARD_CHARACTER}') can not be used in DictPath's")
        # Override type annotation from super class
        self.key: NormalValue

    def get_element(self, container: object, construct: bool = False) -> object:
        elements = WildInDict.get_elements(self, container)

        if not elements and construct:
            if self._validate_container(container):
                container[self.key.value] = {}
                return container[self.key.value]
            else:
                raise ContainerStructureException(f"{container} is not a Dict")

        if len(elements) != 1:
            raise KeyError(f"Found no or multiple items matching {self.to_str()} in {container}: {elements}")

        return elements[0]

    def set_element(self, container: object, value: object, construct: bool = True) -> None:
        if self._validate_container(container):
            container[self.key.value] = value
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def get_path_sections(self) -> Sequence[DictPath]:
        return [self]

    def get_key(self) -> str:
        return self.key.value

    def remove(self, container: object) -> None:
        if self._validate_container(container):
            for key in list(container.keys()):
                if self.key.matches(key):
                    del container[key]
        else:
            raise ContainerStructureException(f"{container} is not a Dict")


@stable_api
class KeyedList(DictPath, WildKeyedList):
    """
    Find a specific item in a list, based on a key-value pair.
    The list is in a dictionary itself.

    The string representation of the following path element is `relation[key_attribute=key_value]`

    e.g.::

        KeyedList("relation","key_attribute","key_value").get_element(
        {
            "relation":[
                {
                    "key_attribute":"key_value",
                    "other_attribute":"other_value"
                },
                {
                    "key_attribute":"other_value"
                }
            [
        })

    will return::

        {
            "key_attribute":"key_value",
            "other_attribute":"other_value"
        }

    """

    def __init__(
        self, relation: str, key_value_pairs: Union[str, Sequence[tuple[str, str]]], key_value: Optional[str] = None
    ) -> None:
        if isinstance(key_value_pairs, str):
            assert key_value is not None
            WildKeyedList.__init__(self, relation, key_value_pairs, key_value)
        else:
            assert key_value is None
            WildKeyedList.__init__(self, relation, key_value_pairs)
        # Override type annotation from super class
        self.key_value_pairs: Sequence[tuple[NormalValue, Union[NormalValue, NullValue]]]

    @classmethod
    def _parse_key(cls, key: str) -> NormalValue:
        result: Union[NormalValue, WildCardValue] = super()._parse_key(key)
        if isinstance(result, WildCardValue):
            raise ValueError(f"The Wildcard ('{WildCardValue.WILDCARD_CHARACTER}') can not be used in DictPath's")
        return result

    @classmethod
    def _parse_value(cls, value: str) -> DictPathValue:
        """
        Parse a value string into the corresponding dict path object.
        """
        result: DictPathValue = super()._parse_value(value)
        if isinstance(result, WildCardValue):
            raise ValueError(f"The Wildcard ('{WildCardValue.WILDCARD_CHARACTER}') can not be used in DictPath's")
        return result

    def get_element(self, container: object, construct: bool = False) -> object:
        found = WildKeyedList.get_elements(self, container)

        if not found and construct:
            outer = self._validate_outer_container(container)
            if self.relation.value not in outer:
                outer[self.relation.value] = []
            the_list = self._validate_inner_container(outer[self.relation.value])
            new_dict: dict[Optional[str], Optional[str]] = {key.value: value.value for key, value in self.key_value_pairs}
            the_list.append(new_dict)
            return new_dict

        if len(found) != 1:
            raise KeyError(f"Found no or multiple items matching {self.to_str()} in {container}: {found}")
        return found[0]

    def set_element(self, container: object, value: object, construct: bool = True) -> None:
        outer: dict[object, object] = self._validate_outer_container(container)
        try:
            inner = outer[self.relation.value]
        except KeyError:
            inner = []
            outer[self.relation.value] = inner

        the_list: list[object] = self._validate_inner_container(inner)
        try:
            element_to_be_replaced: object = self.get_element(container, construct=False)
        except KeyError:
            the_list.append(value)
        else:
            index = the_list.index(element_to_be_replaced)
            the_list[index] = value

    def get_path_sections(self) -> Sequence[DictPath]:
        return [self]

    def get_key(self) -> str:
        return self.relation.value

    def remove(self, container: object) -> None:
        outer = self._validate_outer_container(container)
        try:
            inner = outer[self.relation.value]
        except KeyError:
            return
        the_list = self._validate_inner_container(inner)
        outer[self.relation.value] = [
            dct
            for dct in the_list
            if not isinstance(dct, dict)
            or not all(any(key.matches(k) and value.matches(v) for k, v in dct.items()) for key, value in self.key_value_pairs)
        ]


@stable_api
class ComposedPath(DictPath, WildComposedPath):
    """
    A path composed of multiple elements, separated by "."
    """

    element_types: Sequence[type[DictPath]] = [InDict, KeyedList]

    def __init__(self, path_str: Optional[str] = None, path: Optional[Sequence[DictPath]] = None) -> None:
        WildComposedPath.__init__(self, path_str, path)
        self.expanded_path: Sequence[DictPath]

    def get_element(self, container: object, construct: bool = False) -> object:
        elements = WildComposedPath.get_elements(self, container)

        if not elements and construct:
            element = container
            for item in self.get_path_sections():
                element = item.get_element(element, True)

            return element

        if len(elements) != 1:
            raise KeyError(f"Found no or multiple items matching {self.to_str()} in {container}: {elements}")

        return elements[0]

    def set_element(self, container: object, value: object, construct: bool = True) -> None:
        for item in self.get_path_sections()[:-1]:
            container = item.get_element(container, construct=construct)

        self.get_path_sections()[-1].set_element(container, value)

    def get_path_sections(self) -> Sequence[DictPath]:
        return self.expanded_path

    def get_key(self) -> str:
        raise NotImplementedError("Method get_key() not supported on a ComposedPath")

    def remove(self, container: object) -> None:
        for item in self.get_path_sections()[:-1]:
            try:
                container = item.get_element(container, construct=False)
            except KeyError:
                return

        self.get_path_sections()[-1].remove(container)


@stable_api
class NullPath(DictPath, WildNullPath):
    """
    A DictPath with no length

    (i.e. return the container itself)
    """

    def get_element(self, container: object, construct: bool = False) -> dict[object, object]:
        if self._validate_container(container):
            return container
        else:
            raise ContainerStructureException(f"{container} is not a Dict")

    def set_element(self, container: object, value: object, construct: bool = True) -> None:
        if not self._validate_container(container):
            raise ContainerStructureException(f"Argument container is not a Dict: {container}")
        if not self._validate_container(value):
            raise ContainerStructureException(f"Argument value is not a Dict: {container}")
        assert isinstance(container, dict)
        assert isinstance(value, dict)
        container.clear()
        for key, value in value.items():
            container[key] = value

    @classmethod
    def parse(cls, inp: str) -> None:
        raise NotImplementedError("NullPath is not intended to be parseable, it should only be used programmatically.")

    def get_key(self) -> str:
        raise NotImplementedError("Method get_key() is not supported on a NullPath")

    def remove(self, container: object) -> None:
        raise NotImplementedError("Method remove() is not supported on a NullPath")


@stable_api
def to_wild_path(inp: str) -> WildDictPath:
    """
    Convert a string to a WildDictPath

    :raises InvalidPathException: the path is not valid
    """
    if inp == ".":
        return WildNullPath()
    if inp.startswith("."):
        # A leading dot represents the entire container
        inp = inp[1:]
    try:
        return WildComposedPath(path_str=inp)
    except ValueError as e:
        raise InvalidPathException(str(e))


@stable_api
def to_path(inp: str) -> DictPath:
    """
    Convert a string to a DictPath

    :raises InvalidPathException: the path is not valid
    """
    if inp == ".":
        return NullPath()
    if inp.startswith("."):
        # A leading dot represents the entire container
        inp = inp[1:]
    try:
        return ComposedPath(path_str=inp)
    except ValueError as e:
        raise InvalidPathException(str(e))



"""
    Copyright 2024 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    UPDATE public.parameter
    SET expires=(
        CASE
            WHEN resource_id IS NOT NULL AND resource_id != '' AND expires IS NULL
            -- It's a fact and facts expire by default.
            THEN TRUE
            WHEN resource_id IS NULL OR resource_id = ''
            -- It's a parameter. Parameters never expire.
            THEN FALSE
            -- Keep current value
            ELSE expires
        END
    );

     -- change the type of the 'expires' column to disallow null values

     ALTER TABLE public.parameter
     ALTER COLUMN expires SET NOT NULL;
     """

    await connection.execute(schema)



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os

from pkg_resources import Requirement

from inmanta.file_parser import RequirementsTxtParser


def test_requirements_txt_parser(tmpdir) -> None:
    content = """
        test==1.2.3
        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        split\
ted\
dep
        Capital
    """

    requirements_txt_file = os.path.join(tmpdir, "requirements.txt")
    with open(requirements_txt_file, "w", encoding="utf-8") as fd:
        fd.write(content)

    expected_requirements = ["test==1.2.3", "other-dep~=2.0.0", "third-dep<5.0.0", "splitteddep", "Capital"]
    requirements: list[Requirement] = RequirementsTxtParser().parse(requirements_txt_file)
    assert requirements == [Requirement.parse(r) for r in expected_requirements]
    requirements_as_str = RequirementsTxtParser.parse_requirements_as_strs(requirements_txt_file)
    assert requirements_as_str == expected_requirements

    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="test")
    assert (
        new_content
        == """
        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        splitteddep
        Capital
    """
    )
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="third-dep")
    assert (
        new_content
        == """
        test==1.2.3
        # A comment
        other-dep~=2.0.0
        splitteddep
        Capital
    """
    )
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="splitteddep")
    assert (
        new_content
        == """
        test==1.2.3
        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        Capital
    """
    )
    new_content = RequirementsTxtParser.get_content_with_dep_removed(requirements_txt_file, remove_dep_on_pkg="Capital")
    assert (
        new_content
        == """
        test==1.2.3
        # A comment
        other-dep~=2.0.0
        third-dep<5.0.0 # another comment
        splitteddep
    """
    )



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
from collections import defaultdict

import more_itertools

from inmanta import compiler
from inmanta.ast import Range
from inmanta.compiler import Compiler
from inmanta.execute import scheduler


def test_anchors_basic(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string a = "a"
    string b
end

entity Test2 extends Test:
    foo c
end

Test.relation [0:1] -- Test2.reverse [0:]

typedef foo as string matching /^a+$/

a = Test(b="xx")
z = a.relation
u = a.b

implementation a for Test:

end

implement Test using a
""",
        autostd=False,
    )
    anchormap = compiler.anchormap()

    assert len(anchormap) == 9

    checkmap = {(r.lnr, r.start_char, r.end_char): t.location.lnr for r, t in anchormap}

    def verify_anchor(flnr, s, e, tolnr):
        assert checkmap[(flnr, s, e)] == tolnr

    for f, t in sorted(anchormap, key=lambda x: x[0].lnr):
        print("%s:%d -> %s" % (f, f.end_char, t))
    verify_anchor(7, 22, 26, 2)
    verify_anchor(8, 5, 8, 13)
    verify_anchor(11, 1, 5, 2)
    verify_anchor(11, 24, 29, 7)
    verify_anchor(15, 5, 9, 2)
    verify_anchor(15, 10, 11, 4)
    verify_anchor(19, 22, 26, 2)
    verify_anchor(23, 11, 15, 2)
    verify_anchor(23, 22, 23, 19)


def test_anchors_basic_old(snippetcompiler):
    """
    this test verify that the old path to generate the Anchormap still works. This ensure that we remain
    backward compatible with old Language servers.
    """
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    string a = "a"
    string b
end

entity Test2 extends Test:
    foo c
end

Test.relation [0:1] -- Test2.reverse [0:]

typedef foo as string matching /^a+$/

a = Test(b="xx")
z = a.relation
u = a.b

implementation a for Test:

end

implement Test using a
""",
        autostd=False,
    )
    compiler = Compiler()
    (statements, blocks) = compiler.compile()
    sched = scheduler.Scheduler()
    anchormap = sched.anchormap(compiler, statements, blocks)

    assert len(anchormap) == 9
    assert all(isinstance(item[0], Range) and isinstance(item[1], Range) for item in anchormap)

    checkmap = {(r.lnr, r.start_char, r.end_char): t.lnr for r, t in anchormap}

    def verify_anchor(flnr, s, e, tolnr):
        assert checkmap[(flnr, s, e)] == tolnr

    for f, t in sorted(anchormap, key=lambda x: x[0].lnr):
        print("%s:%d -> %s" % (f, f.end_char, t))
    verify_anchor(7, 22, 26, 2)
    verify_anchor(8, 5, 8, 13)
    verify_anchor(11, 1, 5, 2)
    verify_anchor(11, 24, 29, 7)
    verify_anchor(15, 5, 9, 2)
    verify_anchor(15, 10, 11, 4)
    verify_anchor(19, 22, 26, 2)
    verify_anchor(23, 11, 15, 2)
    verify_anchor(23, 22, 23, 19)


def test_anchors_two(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    list a = ["a"]
    dict b
end

a = Test(b={})
z = a.a
u = a.b

implementation a for Test:

end

implement Test using a
""",
        autostd=False,
    )
    anchormap = compiler.anchormap()

    assert len(anchormap) == 5

    checkmap = {(r.lnr, r.start_char, r.end_char): t.location.lnr for r, t in anchormap}

    def verify_anchor(flnr, s, e, tolnr):
        assert checkmap[(flnr, s, e)] == tolnr

    for f, t in anchormap:
        print("%s:%d -> %s" % (f, f.end_char, t))
    verify_anchor(7, 5, 9, 2)
    verify_anchor(7, 10, 11, 4)
    verify_anchor(11, 22, 26, 2)
    verify_anchor(15, 22, 23, 11)
    verify_anchor(15, 11, 15, 2)


def test_anchors_plugin(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import tests

l = tests::length("Hello World!")
        """
    )
    anchormap = compiler.anchormap()
    location: Range
    resolves_to: Range
    location, resolves_to = more_itertools.one(
        (location, resolves_to)
        for location, resolves_to in anchormap
        if location.file == os.path.join(snippetcompiler.project_dir, "main.cf")
    )
    assert location.lnr == 4
    assert location.start_char == 5
    assert location.end_lnr == 4
    assert location.end_char == 18
    assert resolves_to.location.file == os.path.join(snippetcompiler.modules_dir, "tests", "plugins", "__init__.py")
    assert resolves_to.location.lnr == 14


def test_get_types_and_scopes(snippetcompiler):
    """
    Test the get_types_and_scopes() entrypoint of the compiler.
    """
    snippetcompiler.setup_for_snippet(
        """
    entity Test:
        string a = "a"
        string b
    end

    entity Test2 extends Test:
        foo c
    end

    Test.relation [0:1] -- Test2.reverse [0:]

    typedef foo as string matching /^a+$/

    a = Test(b="xx")
    z = a.relation
    u = a.b

    implementation a for Test:

    end

    implement Test using a

    """
    )

    (types, scopes) = compiler.get_types_and_scopes()

    # Verify types
    namespace_to_type_name = defaultdict(list)
    for type_name in types.keys():
        namespace = type_name.split("::")[0]
        namespace_to_type_name[namespace].append(type_name)

    assert len(namespace_to_type_name) == 2
    assert "__config__" in namespace_to_type_name
    assert "std" in namespace_to_type_name

    # Assert types in namespace __config__
    expected_types_in_config_ns = [
        "__config__::Test",
        "__config__::Test2",
        "__config__::foo",
        "__config__::a",
    ]
    assert sorted(namespace_to_type_name["__config__"]) == sorted(expected_types_in_config_ns)

    # Assert types in namespace std
    types_in_std_ns = namespace_to_type_name["std"]
    assert len(types_in_std_ns) > 1
    assert "std::Entity" in types_in_std_ns

    # Verify scopes
    assert scopes.get_name() == "__root__"
    assert sorted([scope.get_name() for scope in scopes.children()]) == sorted(["__config__", "std"])


def test_anchors_with_docs(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import tests

l = tests::length("Hello World!") # has a docstring
m = tests::empty("Hello World!") # has no docstring

entity Test:
    \"\"\"this is a test entity\"\"\"
end

entity Test_no_doc:
end

a = Test()
b = Test_no_doc()

implementation a for Test:
end

implementation b for Test_no_doc:
end

implement Test using a
implement Test_no_doc using b
""",
        autostd=False,
    )
    anchormap = compiler.anchormap()

    assert len(anchormap) == 10

    checkmap = {(r.lnr, r.start_char, r.end_char): t.docstring for r, t in anchormap}

    def verify_anchor(flnr, s, e, docs):
        assert checkmap[(flnr, s, e)] == docs

    for f, t in sorted(anchormap, key=lambda x: x[0].lnr):
        print("%s:%d -> %s docstring: %s" % (f, f.end_char, t, t.docstring))
    verify_anchor(4, 5, 18, "returns the length of the string")
    verify_anchor(5, 5, 17, None)
    verify_anchor(14, 5, 9, "this is a test entity")
    verify_anchor(15, 5, 16, None)
    verify_anchor(17, 22, 26, "this is a test entity")
    verify_anchor(20, 22, 33, None)
    verify_anchor(23, 22, 23, None)
    verify_anchor(23, 11, 15, "this is a test entity")
    verify_anchor(24, 29, 30, None)
    verify_anchor(24, 11, 22, None)


def test_constructor_with_inferred_namespace(snippetcompiler):
    """
    Test that the anchor for a constructor for an entity with an inferred namespace is correctly added to the anchormap
    The test checks if the anchormap correctly reflects the relationship between the
    source range (where the entity is instantiated: line 9 in main.cf) and the target range (where the
    entity is defined: line 1 in the _init.cf file of the tests module)
    """

    module: str = "tests"
    target_path = os.path.join(os.path.dirname(__file__), "data", "modules", module, "model", "_init.cf")

    snippetcompiler.setup_for_snippet(
        """
    import mod1
    import tests
    entity A:
    end

    A.mytest [1] -- tests::Test [1]

    A(mytest = Test())
    """,
        autostd=False,
    )

    compiler = Compiler()
    (statements, blocks) = compiler.compile()
    sched = scheduler.Scheduler()

    anchormap = sched.anchormap(compiler, statements, blocks)
    assert len(anchormap) == 5
    range_source = Range(os.path.join(snippetcompiler.project_dir, "main.cf"), 9, 16, 9, 20)
    range_target = Range(target_path, 1, 8, 1, 12)

    assert (range_source, range_target) in anchormap


def test_constructor_renamed_namespace(snippetcompiler):
    """
    Test that the anchor for a constructor with `import a as b` works
    """

    module: str = "tests"
    target_path = os.path.join(os.path.dirname(__file__), "data", "modules", module, "model", "subpack", "submod.cf")

    snippetcompiler.setup_for_snippet(
        """
    import mod1
    import tests::subpack::submod as t
    entity A:
    end

    A.mytest [1] -- t::Test [1]

    A(mytest = t::Test())
    """,
        autostd=False,
    )

    compiler = Compiler()
    (statements, blocks) = compiler.compile()
    sched = scheduler.Scheduler()

    anchormap = sched.anchormap(compiler, statements, blocks)
    assert len(anchormap) == 5
    range_source = Range(os.path.join(snippetcompiler.project_dir, "main.cf"), 9, 16, 9, 23)
    range_target = Range(target_path, 1, 8, 1, 12)

    assert (range_source, range_target) in anchormap



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import pkgutil
import re
from collections.abc import Coroutine
from types import ModuleType
from typing import Any, Callable, Optional

from asyncpg import Connection, UndefinedTableError
from asyncpg.protocol import Record

# Name of core schema in the DB schema verions
CORE_SCHEMA_NAME = "core"

LOGGER = logging.getLogger(__name__)

SCHEMA_VERSION_TABLE = "schemamanager"

create_schemamanager = """
-- Table: public.schemamanager
CREATE TABLE IF NOT EXISTS public.schemamanager (
    name varchar PRIMARY KEY,
    legacy_version integer,
    installed_versions integer[]
);
"""


class TableNotFound(Exception):
    """Raised when a table is not found in the database"""


class ColumnNotFound(Exception):
    """Raised when a column is not found in the database"""


class Version:
    """Internal representation of a version"""

    def __init__(self, name: str, function: Callable[[Connection], Coroutine[Any, Any, None]]):
        self.name = name
        self.function = function
        self.version = self.parse(name)

    @classmethod
    def parse(cls, name: str) -> int:
        return int(name[1:])


class DBSchema:
    """
    Schema Manager, ensures the schema is up to date.

    Concurrent updates are safe
    """

    def __init__(self, name: str, package: ModuleType, connection: Connection) -> None:
        """

        :param name: unique name for this schema, best equal to extension name and used as prefix for all table names
        :param package: a python package, containing modules with name v%(version)s.py.
            Each module contains a method `async def update(connection: asyncpg.connection) -> None:`
        :param connection: asyncpg connection
        """
        self.name = name
        self.package = package
        self.connection = connection
        self.logger = LOGGER.getChild(f"schema:{self.name}")

    async def ensure_db_schema(self) -> None:
        await self.ensure_self_update()
        await self._update_db_schema()

    async def ensure_self_update(self) -> None:
        """
        Ensures the table exists and is up to date with respect to the current schema.
        """
        self.logger.info("Creating schema version table")
        await self.connection.execute(create_schemamanager)

    async def _update_db_schema(self, update_functions: Optional[list[Version]] = None) -> None:
        """
        Main update function

        Wrapped in transaction, that holds a lock on the schemamanager table.
        When a version update fails, the whole transaction is rolled back.

        :param update_functions: allows overriding the available update functions, for example for testing purposes.
        """
        update_functions = (
            sorted(update_functions, key=lambda x: x.version) if update_functions is not None else self._get_update_functions()
        )
        async with self.connection.transaction():
            # get lock
            await self.connection.execute(f"LOCK TABLE {SCHEMA_VERSION_TABLE} IN ACCESS EXCLUSIVE MODE")
            # get current version again, in transaction this time
            try:
                installed_versions: set[int] = await self.get_installed_versions()
            except TableNotFound:
                self.logger.exception("Schemamanager table disappeared, should not occur.")
                raise
            # get relevant updates
            updates = [v for v in update_functions if v.version not in installed_versions]
            for version in updates:
                try:
                    # actual update sequence
                    self.logger.info("Updating database schema to version %d", version.version)
                    update_function = version.function
                    await update_function(self.connection)
                    await self.set_installed_version(version.version)
                    # inform asyncpg of the type change so it knows to refresh its caches
                    await self.connection.reload_schema_state()
                except Exception:
                    self.logger.exception(
                        "Database schema update for version %d failed. Rolling back all updates.",
                        version.version,
                    )
                    # propagate exception => roll back transaction
                    raise

    async def get_installed_versions(self) -> set[int]:
        """
        Returns the set of all versions that have been installed.

        :raises TableNotFound:
        """
        versions: Optional[Record] = None
        try:
            versions = await self.connection.fetchrow(
                f"select installed_versions from {SCHEMA_VERSION_TABLE} where name=$1", self.name
            )
        except UndefinedTableError as e:
            raise TableNotFound() from e
        if versions is None or versions["installed_versions"] is None:
            return set()
        return set(versions["installed_versions"])

    async def set_installed_version(self, version: int) -> None:
        """
        Adds a version to the installed versions column.
        """
        await self.connection.execute(
            f"""
            INSERT INTO {SCHEMA_VERSION_TABLE} (name, installed_versions)
            VALUES ($1, $2) ON CONFLICT (name) DO UPDATE
            SET installed_versions = {SCHEMA_VERSION_TABLE}.installed_versions || excluded.installed_versions
            """,
            self.name,
            {version},
        )

    def _get_update_functions(self) -> list[Version]:
        module_names = [modname for _, modname, ispkg in pkgutil.iter_modules(self.package.__path__) if not ispkg]

        def get_modules(mod_name: str) -> tuple[str, ModuleType]:
            fq_module_name = self.package.__name__ + "." + mod_name
            return mod_name, __import__(fq_module_name, fromlist=["update"])

        def make_version(mod_name: str, module: ModuleType) -> Version:
            update_function = module.update
            return Version(mod_name, update_function)

        def disabled(module: ModuleType) -> bool:
            try:
                return module.DISABLED
            except AttributeError:
                return False

        pattern = re.compile("^v[0-9]+$")

        filtered_module_names = []
        for module_name in module_names:
            if not pattern.match(module_name):
                LOGGER.warning(
                    f"Database schema version file name {module_name} "
                    f"doesn't match the expected pattern: v<version_number>.py, skipping it"
                )
            else:
                filtered_module_names.append(module_name)

        modules_with_names = [get_modules(mod_name) for mod_name in filtered_module_names]
        filtered_modules = [(module_name, module) for module_name, module in modules_with_names if not disabled(module)]

        version = [make_version(name, mod) for name, mod in filtered_modules]

        return sorted(version, key=lambda x: x.version)
